/* Include the controller definition */
#include "footbot_cleaning_robots.h"
/* Function definitions for XML parsing */
#include <argos3/core/utility/configuration/argos_configuration.h>
/* 2D vector definition */
#include <argos3/core/utility/math/vector2.h>
/* Logging */
#include <argos3/core/utility/logging/argos_log.h>

/**********************************************************************************/
CFootBotCleaningRobots::CFootBotCleaningRobots() :
	m_pcWheels(NULL), m_pcLEDs(NULL), m_pcProximity(NULL), m_pcGround(NULL), m_pcLight(NULL),  
	m_RNG(NULL), m_cProbRange(0, 100), time_counter(0) {}
/**********************************************************************************/
CFootBotCleaningRobots::SDiffusionParams::SDiffusionParams() :
   GoStraightAngleRange(CRadians(-1.0f), CRadians(1.0f)) {}
/**********************************************************************************/
CFootBotCleaningRobots::SRobotInfo::SRobotInfo() :
   call_substituted_robot(false),add_me_to_deactivated_counter(false),add_me_to_activated_counter(false) {}
/**********************************************************************************/
CFootBotCleaningRobots::SBirthDeathParams::SBirthDeathParams() :
   m_fMu(0.0f), m_fLambda(0.0f) {}
/**********************************************************************************/
void CFootBotCleaningRobots::SDiffusionParams::Init(TConfigurationNode& t_node) {
   try {
      CRange<CDegrees> cGoStraightAngleRangeDegrees(CDegrees(-10.0f), CDegrees(10.0f));
      GetNodeAttribute(t_node, "go_straight_angle_range", cGoStraightAngleRangeDegrees);
      GoStraightAngleRange.Set(ToRadians(cGoStraightAngleRangeDegrees.GetMin()),
                               ToRadians(cGoStraightAngleRangeDegrees.GetMax()));
      GetNodeAttribute(t_node, "delta", Delta);
   }
   catch(CARGoSException& ex) {
      THROW_ARGOSEXCEPTION_NESTED("Error initializing controller diffusion parameters.", ex);
   }
}
/**********************************************************************************/
void CFootBotCleaningRobots::SWheelTurningParams::Init(TConfigurationNode& t_node) {
   try {
      CDegrees cAngle;
      GetNodeAttribute(t_node, "hard_turn_angle_threshold", cAngle);
      HardTurnOnAngleThreshold = ToRadians(cAngle);
      GetNodeAttribute(t_node, "soft_turn_angle_threshold", cAngle);
      SoftTurnOnAngleThreshold = ToRadians(cAngle);
      GetNodeAttribute(t_node, "no_turn_angle_threshold", cAngle);
      NoTurnAngleThreshold = ToRadians(cAngle);
      GetNodeAttribute(t_node, "max_speed", MaxSpeed);
   }
   catch(CARGoSException& ex) {
      THROW_ARGOSEXCEPTION_NESTED("Error initializing controller wheel turning parameters.", ex);
   }
}
/**********************************************************************************/
void CFootBotCleaningRobots::SBirthDeathParams::Init(TConfigurationNode& t_node) {
   try {
      /* death rate */
      GetNodeAttribute(t_node, "mu", m_fMu);
      /* birth rate */
      GetNodeAttribute(t_node, "lambda", m_fLambda);
   }
   catch(CARGoSException& ex) {
      THROW_ARGOSEXCEPTION_NESTED("Error initializing birth-death parameters.", ex);
   }
}
/**********************************************************************************/

void CFootBotCleaningRobots::Init(TConfigurationNode& t_node) {
	m_pcWheels = GetActuator<CCI_DifferentialSteeringActuator>("differential_steering");
	m_pcLEDs = GetActuator<CCI_LEDsActuator>("leds");
	m_pcProximity = m_pcProximity = GetSensor  <CCI_FootBotProximitySensor>("footbot_proximity");
	m_pcGround    = GetSensor  <CCI_FootBotMotorGroundSensor>("footbot_motor_ground" );
    m_pcLight     = GetSensor  <CCI_FootBotLightSensor          >("footbot_light"        );
    //random generator 
	m_RNG = CRandom::CreateRNG("argos");
	/*
	 * Parse the XML
	 */
	
	/* Wheel turning */
      m_sWheelTurningParams.Init(GetNode(t_node, "wheel_turning"));
      /* Diffusion algorithm */
      m_sDiffusionParams.Init(GetNode(t_node, "diffusion"));
    /* Birth-death process */
      m_SBirthDeathParams.Init(GetNode(t_node, "birth_death"));

	m_sRobotInfo.sRobotState = SRobotInfo::CLEANING;
	m_sRobotInfo.sRobotBirthDeathState = SRobotInfo::ACTIVE;
	
}

/**********************************************************************************/

/*
 * The vector returned by CalculateVectorToLight() points to the light.
 */
CVector2 CFootBotCleaningRobots::CalculateVectorToLight() {
   /* Get readings from light sensor */
   const CCI_FootBotLightSensor::TReadings& tLightReads = m_pcLight->GetReadings();
   /* Sum them together */
   CVector2 cAccumulator;
   for(size_t i = 0; i < tLightReads.size(); ++i) {
      cAccumulator += CVector2(tLightReads[i].Value, tLightReads[i].Angle);
   }
   /* If the light was perceived, return the vector */
   if(cAccumulator.Length() > 0.0f) {
      return CVector2(1.0f, cAccumulator.Angle());
   }
   /* Otherwise, return zero */
   else {
      return CVector2();
   }
}
/**********************************************************************************/

void CFootBotCleaningRobots::SetWheelSpeedsFromVector(const CVector2& c_heading) {
   /* Get the heading angle from the given vector */
   CRadians cHeadingAngle = c_heading.Angle().SignedNormalize();
   /* Get the length of the heading vector */
   Real fHeadingLength = c_heading.Length();
   /* Clamp the speed so that it's not greater than MaxSpeed */
   Real fBaseAngularWheelSpeed = Min<Real>(fHeadingLength, m_sWheelTurningParams.MaxSpeed);

   /* Turning state switching conditions */
   if(Abs(cHeadingAngle) <= m_sWheelTurningParams.NoTurnAngleThreshold) {
      /* No Turn, heading angle very small */
      m_sWheelTurningParams.TurningMechanism = SWheelTurningParams::NO_TURN;
   }
   else if(Abs(cHeadingAngle) > m_sWheelTurningParams.HardTurnOnAngleThreshold) {
      /* Hard Turn, heading angle very large */
      m_sWheelTurningParams.TurningMechanism = SWheelTurningParams::HARD_TURN;
   }
   else if(m_sWheelTurningParams.TurningMechanism == SWheelTurningParams::NO_TURN &&
           Abs(cHeadingAngle) > m_sWheelTurningParams.SoftTurnOnAngleThreshold) {
      /* Soft Turn, heading angle in between the two cases */
      m_sWheelTurningParams.TurningMechanism = SWheelTurningParams::SOFT_TURN;
   }

   /* Wheel speeds based on current turning state */
   Real fSpeed1, fSpeed2;
   switch(m_sWheelTurningParams.TurningMechanism) {
      case SWheelTurningParams::NO_TURN: {
         /* Just go straight */
         fSpeed1 = fBaseAngularWheelSpeed;
         fSpeed2 = fBaseAngularWheelSpeed;
         break;
      }

      case SWheelTurningParams::SOFT_TURN: {
         /* Both wheels go straight, but one is faster than the other */
         Real fSpeedFactor = (m_sWheelTurningParams.HardTurnOnAngleThreshold - Abs(cHeadingAngle)) / m_sWheelTurningParams.HardTurnOnAngleThreshold;
         fSpeed1 = fBaseAngularWheelSpeed - fBaseAngularWheelSpeed * (1.0 - fSpeedFactor);
         fSpeed2 = fBaseAngularWheelSpeed + fBaseAngularWheelSpeed * (1.0 - fSpeedFactor);
         break;
      }

      case SWheelTurningParams::HARD_TURN: {
         /* Opposite wheel speeds */
         fSpeed1 = -m_sWheelTurningParams.MaxSpeed;
         fSpeed2 =  m_sWheelTurningParams.MaxSpeed;
         break;
      }
   }

   /* Apply the calculated speeds to the appropriate wheels */
   Real fLeftWheelSpeed, fRightWheelSpeed;
   if(cHeadingAngle > CRadians::ZERO) {
      /* Turn Left */
      fLeftWheelSpeed  = fSpeed1;
      fRightWheelSpeed = fSpeed2;
   }
   else {
      /* Turn Right */
      fLeftWheelSpeed  = fSpeed2;
      fRightWheelSpeed = fSpeed1;
   }
   /* Finally, set the wheel speeds */
   m_pcWheels->SetLinearVelocity(fLeftWheelSpeed, fRightWheelSpeed);
}
/**********************************************************************************/

CVector2 CFootBotCleaningRobots::DiffusionVector(bool& b_collision) {
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();
   /* Sum them together */
   CVector2 cDiffusionVector;
   for(size_t i = 0; i < tProxReads.size(); ++i) {
      cDiffusionVector += CVector2(tProxReads[i].Value, tProxReads[i].Angle);
   }
   /* If the angle of the vector is small enough and the closest obstacle
      is far enough, ignore the vector and go straight, otherwise return
      it */
   if(m_sDiffusionParams.GoStraightAngleRange.WithinMinBoundIncludedMaxBoundIncluded(cDiffusionVector.Angle()) &&
      cDiffusionVector.Length() < m_sDiffusionParams.Delta ) {
      b_collision = false;
      return CVector2::X;
   }
   else {
      b_collision = true;
      cDiffusionVector.Normalize();
      return -cDiffusionVector;
   }
}
/********************************************************************************/
/**
 * Control step of the controller
 */
void CFootBotCleaningRobots::ControlStep() { 

	time_counter = time_counter+1;
 if ((m_sRobotInfo.sRobotState == SRobotInfo::CLEANING)||(m_sRobotInfo.sRobotState == SRobotInfo::AVOIDING)){
 /* The robot is active when it is in any of the CLEANING or the AVODING states */
 m_sRobotInfo.sRobotBirthDeathState = SRobotInfo::ACTIVE;
	/*
	 * Check the probability by a working robot to leave the task (e.g. defect) 
	*/
	if(time_counter==10){
	time_counter = 0;
	UInt32 defect_prob = m_RNG->Uniform(m_cProbRange);
	if (defect_prob <= m_SBirthDeathParams.m_fMu){
		m_sRobotInfo.sRobotBirthDeathState = SRobotInfo::DEACTIVATED;
		m_sRobotInfo.add_me_to_deactivated_counter = true;
		/* check the probability by the leaving robot to activate an alternative one */
		UInt32 activation_prob = m_RNG->Uniform(m_cProbRange);
		if (activation_prob <= m_SBirthDeathParams.m_fLambda){
		/* Robot should be substituted */
		m_sRobotInfo.call_substituted_robot = true;
		m_sRobotInfo.sRobotBirthDeathState = SRobotInfo::ACTIVATED;
		m_sRobotInfo.add_me_to_activated_counter=true;
		} else if(activation_prob > m_SBirthDeathParams.m_fLambda)
		{
			m_sRobotInfo.sRobotState = SRobotInfo::DEFECTED;
			m_pcLEDs->SetAllColors(CColor::RED);
			m_pcWheels->SetLinearVelocity(0, 0);
		}		
	}  
  }
  }
	 if (m_sRobotInfo.sRobotState != SRobotInfo::DEFECTED){
	/* Read sensed data from the Proximity sensor */
  	 const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();
	/* Read sensed data from the ground sensor */
  	 const CCI_FootBotMotorGroundSensor::TReadings& tGroundReads = m_pcGround->GetReadings();
  	 /* check the ground readings to verify if the robot is on the maintenance area */
  	 if(tGroundReads[2].Value > 0.25f &&
     	 tGroundReads[2].Value < 0.75f &&
     	 tGroundReads[3].Value > 0.25f &&
     	 tGroundReads[3].Value < 0.75f &&
     	 ! (m_sRobotInfo.sRobotState == SRobotInfo::DEFECTED)) { 
      	SetWheelSpeedsFromVector(-0.9f * m_sWheelTurningParams.MaxSpeed * CalculateVectorToLight());  
   } else {
   /* Get the diffusion vector to perform exploration with obstacle avoidance */
      bool bCollision;
      CVector2 cDiffusion = DiffusionVector(bCollision);
      if(bCollision){
       m_pcLEDs->SetAllColors(CColor::WHITE);
       m_sRobotInfo.sRobotState = SRobotInfo::AVOIDING;
      } else{
       m_pcLEDs->SetAllColors(CColor::GREEN);
       m_sRobotInfo.sRobotState = SRobotInfo::CLEANING;
      }
      SetWheelSpeedsFromVector(m_sWheelTurningParams.MaxSpeed * cDiffusion);
		}
	}
	
}
/**********************************************************************************/
/*
 * This statement notifies ARGoS of the existence of the controller.
 * It binds the class passed as first argument to the string passed as second argument.
 * The string is then usable in the XML configuration file to refer to this controller.
 * When ARGoS reads that string in the XML file, it knows which controller class to instantiate.
 * See also the XML configuration files for an example of how this is used.
 */
REGISTER_CONTROLLER(CFootBotCleaningRobots, "footbot_cleaning_robots_controller")
