/*
 * AUTHOR: Yara Khaluf <yara@upb.de>
 *
 * An example diffusion controller for the foot-bot.
 *
 * This controller makes the robots behave as gas particles. The robots
 * go straight until they get close enough to another robot, in which
 * case they turn, loosely simulating an elastic collision. The net effect
 * is that over time the robots diffuse in the environment.
 *
 * The controller uses the proximity sensor to detect obstacles and the
 * wheels to move the robot around. The LEDs are used to visually show to
 * the user the angle where closest obstacle was detected.
 *
 * This controller is meant to be used with the XML files:
 *    xml/diffusion_1.xml
 *    xml/diffusion_10.xml
 */

#ifndef FOOTBOT_CLEANING_ROBOTS_H
#define FOOTBOT_CLEANING_ROBOTS_H

/*
 * Include some necessary headers.
 */
/* Definition of the CCI_Controller class. */
#include <argos3/core/control_interface/ci_controller.h>
/* Definition of the differential steering actuator */
#include <argos3/plugins/robots/generic/control_interface/ci_differential_steering_actuator.h>
/* Definition of the LEDs actuator */
#include <argos3/plugins/robots/generic/control_interface/ci_leds_actuator.h>
/* Definition of the foot-bot proximity sensor */
#include <argos3/plugins/robots/foot-bot/control_interface/ci_footbot_proximity_sensor.h>
/* Definition of the foot-bot motor ground sensor */
#include <argos3/plugins/robots/foot-bot/control_interface/ci_footbot_motor_ground_sensor.h>
/* Definition of the foot-bot light sensor */
#include <argos3/plugins/robots/foot-bot/control_interface/ci_footbot_light_sensor.h>
/* Definitions for random number generation */
#include <argos3/core/utility/math/rng.h>

/*
 * All the ARGoS stuff in the 'argos' namespace.
 * With this statement, you save typing argos:: every time.
 */
using namespace argos;

/*
 * A controller is simply an implementation of the CCI_Controller class.
 */
class CFootBotCleaningRobots : public CCI_Controller {

public:
	/*
    * The following struct describes the states at which the robot can be.
    */
	struct SRobotInfo{
		enum STATE {CLEANING = 0, AVOIDING, DEFECTED} sRobotState;
		/* This state is used to make statistics in the loop function about the state the robot is concerning the birth-death process*/
		enum BIRTH_DEATH_STATE {ACTIVE = 0, DEACTIVATED, ACTIVATED} sRobotBirthDeathState;
		bool call_substituted_robot;
		bool add_me_to_deactivated_counter;
		bool add_me_to_activated_counter;
		SRobotInfo();
	};
	 /*
    * The following variables are used as parameters for the
    * diffusion algorithm. You can set their value in the <parameters>
    * section of the XML configuration file, under the
    * <controllers><footbot_foraging_controller><parameters><diffusion>
    * section.
    */
   struct SDiffusionParams {
      /*
       * Maximum tolerance for the proximity reading between
       * the robot and the closest obstacle.
       * The proximity reading is 0 when nothing is detected
       * and grows exponentially to 1 when the obstacle is
       * touching the robot.
       */
      Real Delta;
      /* Angle tolerance range to go straight. */
      CRange<CRadians> GoStraightAngleRange;

      /* Constructor */
      SDiffusionParams();

      /* Parses the XML section for diffusion */
      void Init(TConfigurationNode& t_tree);
   };
	
	 /*
    * The following variables are used as parameters for
    * turning during navigation. You can set their value
    * in the <parameters> section of the XML configuration
    * file, under the
    * <controllers><footbot_foraging_controller><parameters><wheel_turning>
    * section.
    */
   struct SWheelTurningParams {
      /*
       * The turning mechanism.
       * The robot can be in three different turning states.
       */
      enum ETurningMechanism
      {
         NO_TURN = 0, // go straight
         SOFT_TURN,   // both wheels are turning forwards, but at different speeds
         HARD_TURN    // wheels are turning with opposite speeds
      } TurningMechanism;
      /*
       * Angular thresholds to change turning state.
       */
      CRadians HardTurnOnAngleThreshold;
      CRadians SoftTurnOnAngleThreshold;
      CRadians NoTurnAngleThreshold;
      /* Maximum wheel speed */
      Real MaxSpeed;

      void Init(TConfigurationNode& t_tree);
   };
 /*
    * The following variables are used as parameters for
    * the birth-death process. You can set their value
    * in the <parameters> section of the XML configuration
    * file, under the
    * <controllers><footbot_foraging_controller><parameters><birth_death>
    * section.
    */
 struct SBirthDeathParams {
      /* Death rate */
      Real m_fMu;
      /* Birth rate */
       Real m_fLambda;

      /* Constructor */
      SBirthDeathParams();

      /* Parses the XML section for birth-death */
      void Init(TConfigurationNode& t_tree);
   };
public:

   /* Class constructor. */
	CFootBotCleaningRobots();

   /* Class destructor. */
   virtual ~CFootBotCleaningRobots() {}

   /*
    * This function initializes the controller.
    * The 't_node' variable points to the <parameters> section in the XML file
    * in the <controllers><footbot_diffusion_controller> section.
    */
   virtual void Init(TConfigurationNode& t_node);

   /*
    * This function is called once every time step.
    * The length of the time step is set in the XML file.
    */
   virtual void ControlStep();
   
  
   /*
    * This function resets the controller to its state right after the Init().
    * It is called when you press the reset button in the GUI.
    * In this example controller there is no need for resetting anything, so
    * the function could have been omitted. It's here just for completeness.
    */
   virtual void Reset() {}

   /*
    * Called to cleanup what done by Init() when the experiment finishes.
    * In this example controller there is no need for clean anything up, so
    * the function could have been omitted. It's here just for completeness.
    */
   virtual void Destroy() {}
   /*
   * get the robot info
   */
   inline SRobotInfo& GetRobotInfo() {
         return m_sRobotInfo;
      }
private:
/*
    * Calculates the vector to the light. Used to perform
    * phototaxis and antiphototaxis.
    */
virtual CVector2 CalculateVectorToLight();
/*
    * Calculates the diffusion vector.
    */
virtual CVector2 DiffusionVector(bool& b_collision);
 /*
    * Gets a direction vector as input and transforms it into wheel
    * actuation.
    */
 virtual void SetWheelSpeedsFromVector(const CVector2& c_heading);
private:

   /* Pointer to the foot-bot wheels actuator */
   CCI_DifferentialSteeringActuator*  m_pcWheels;
   /* Pointer to the foot-bot LEDs actuator */
   CCI_LEDsActuator*  m_pcLEDs;
   /* Pointer to the foot-bot proximity sensor */
   CCI_FootBotProximitySensor* m_pcProximity;
 	/* Pointer to the foot-bot motor ground sensor */
   CCI_FootBotMotorGroundSensor* m_pcGround;
   /* Pointer to the foot-bot light sensor */
   CCI_FootBotLightSensor* m_pcLight;
   CRandom::CRNG* m_RNG;
   /* probability range */
   CRange<UInt32> m_cProbRange; 
    /* The turning parameters */
   SWheelTurningParams m_sWheelTurningParams;
     /* The diffusion parameters */
   SDiffusionParams m_sDiffusionParams;
   /* The robots info structure */
   SRobotInfo m_sRobotInfo;
   /* The birth-death parameters */
   SBirthDeathParams m_SBirthDeathParams;
   size_t time_counter;

};

#endif
